<?php
/* --------------------------------------------------------------
   GambioHubUpdater.inc.php 2017-10-25
   Gambio GmbH
   http://www.gambio.de
   Copyright (c) 2017 Gambio GmbH
   Released under the GNU General Public License (Version 2)
   [http://www.gnu.org/licenses/gpl-2.0.html]
   --------------------------------------------------------------
*/

MainFactory::load_class('HubConnectorUpdaterInterface');

/**
 * Class HubConnectorUpdater
 */
class HubConnectorUpdater implements HubConnectorUpdaterInterface
{
	/**
	 * @var string
	 */
	private $hubConnectorVersion;
	
	/**
	 * @var string
	 */
	private $hubConnectorUpdaterVersion;
	
	/**
	 * @var bool
	 */
	private $downloadBeforeUpdate;
	
	/**
	 * @var bool
	 */
	private $downloadSuccessNotification;
	
	/**
	 * @var array
	 */
	private $downloadedUpdates;
	
	/**
	 * @var string
	 */
	private $gmKeyVersion;
	
	/**
	 * @var string
	 */
	private $gmKeyURL;
	
	/**
	 * @var messageStack_ORIGIN
	 */
	private $messageStack;
	
	/**
	 * @var LanguageTextManager
	 */
	private $textPhrases;
	
	/**
	 * @var string
	 */
	private $backupsDirectory;
	
	/**
	 * @var string
	 */
	private $updatesDirectory;
	
	/**
	 * @var string
	 */
	private $updateServer;
	
	/**
	 * @var array
	 */
	private $updateStatus;
	
	
	/**
	 * HubConnectorUpdater constructor.
	 */
	public function __construct()
	{
		$this->downloadBeforeUpdate        = true;
		$this->downloadSuccessNotification = false;
		$this->gmKeyVersion                = 'HUB_CONNECTOR_UPDATER_VERSION';
		$this->gmKeyURL                    = 'HUB_CONNECTOR_UPDATER_URL';
		$this->hubConnectorVersion         = $this->getHubConnectorVersion();
		$this->hubConnectorUpdaterVersion  = $this->getHubConnectorUpdaterVersion();
		$this->textPhrases                 = MainFactory::create('LanguageTextManager', 'hub_connector_updater',
		                                                         $_SESSION['languages_id']);
		$this->backupsDirectory            = DIR_FS_CATALOG . 'cache' . DIRECTORY_SEPARATOR . 'hubConnectorBackups';
		$this->updatesDirectory            = DIR_FS_CATALOG . 'cache' . DIRECTORY_SEPARATOR . 'hubConnectorUpdates';
		$this->messageStack                = $GLOBALS['messageStack'];
		
		$this->updateServer = gm_get_conf($this->gmKeyURL);
		if(empty($this->updateServer))
		{
			$this->updateServer = 'https://updates.gambio-support.de/check.php';
		}
		
		try
		{
			$this->checkUpdates();
			$this->checkDownloadedUpdates();
		}
		catch(HubConnectorUpdaterServerNotAvailableException $e)
		{
			$this->messageStack->add($e->getMessage(), 'info');
		}
	}
	
	
	/**
	 * Returns the current version of the Hub Connector Updater
	 *
	 * @return string
	 */
	public function getHubConnectorUpdaterVersion()
	{
		if(!empty($this->hubConnectorUpdaterVersion))
		{
			return $this->hubConnectorUpdaterVersion;
		}
		
		$hubConnectorUpdaterVersion = gm_get_conf($this->gmKeyVersion);
		$hubConnectorVersion        = $this->getHubConnectorVersion();
		
		if(empty($hubConnectorUpdaterVersion) && $hubConnectorVersion === '0.0.0')
		{
			$hubConnectorUpdaterVersion = '0.0.0';
			gm_set_conf($this->gmKeyVersion, $hubConnectorUpdaterVersion);
		}
		elseif(empty($hubConnectorUpdaterVersion) && $hubConnectorVersion !== '0.0.0')
		{
			$hubConnectorUpdaterVersion = '1.0.0';
			gm_set_conf($this->gmKeyVersion, $hubConnectorUpdaterVersion);
		}
		elseif($hubConnectorUpdaterVersion === '0.0.0' && $hubConnectorVersion !== '0.0.0')
		{
			$hubConnectorUpdaterVersion = '1.0.0';
			gm_set_conf($this->gmKeyVersion, $hubConnectorUpdaterVersion);
		}
		
		return $hubConnectorUpdaterVersion;
	}
	
	
	/**
	 * Returns the current version of the Hub Connector
	 *
	 * @return string
	 */
	public function getHubConnectorVersion()
	{
		if(!empty($this->hubConnectorVersion))
		{
			return $this->hubConnectorVersion;
		}
		
		$db       = StaticGXCoreLoader::getDatabaseQueryBuilder();
		$versions = $db->select('version')
		               ->from('version_history')
		               ->where('name LIKE "Gambio Hub%" and name NOT LIKE "Gambio Hub Connector Updater%"')
		               ->order_by('version', 'DESC')
		               ->get()
		               ->result();
		
		if(count($versions) > 0 && !empty($versions[0]->version))
		{
			return $versions[0]->version;
		}
		
		return '0.0.0';
	}
	
	
	/**
	 * Returns the current value of the "download before update" option.
	 *
	 * @return bool
	 */
	public function getDownloadBeforeUpdate()
	{
		return $this->downloadBeforeUpdate;
	}
	
	
	/**
	 * Returns the update status as an array. The array contains the keys body and info, which contains the body and
	 * header information of the curl request that was made.
	 *
	 * @return array
	 */
	public function getUpdatesStatus()
	{
		return $this->updateStatus;
	}
	
	
	/**
	 * Returns an arraywith all update versions which are already downloaded.
	 *
	 * @return array
	 */
	public function getDownloadedUpdates()
	{
		return $this->downloadedUpdates;
	}
	
	
	/**
	 * Starts the download of all available updates.
	 *
	 * @throws \HubConnectorUpdaterDownloadFailedException
	 * @throws \HubConnectorUpdaterNoUpdateAvailableException
	 */
	public function downloadUpdates()
	{
		$this->deleteUpdatesDirectory();
		
		if(count($this->updateStatus['body']['updates']) === 0)
		{
			throw new HubConnectorUpdaterNoUpdateAvailableException($this->textPhrases->get_text('NO_UPDATES'));
		}
		
		if(!class_exists('ZipArchive'))
		{
			$this->messageStack->add_session($this->textPhrases->get_text('ERROR_ZIPCLASS_MISSING'), 'notice');
			
			foreach($this->updateStatus['body']['updates'] as $update)
			{
				try
				{
					$this->downloadUpdateNoZip($update);
				}
				catch(Exception $e)
				{
					if($e instanceof HubConnectorUpdaterUpdateNotAvailableException === false
					   && $e instanceof HubConnectorUpdaterMD5HashException === false
					   && $e instanceof HubConnectorUpdaterNoPermissionException === false
					)
					{
						# catch only own exceptions
						throw $e;
					}
					elseif($e instanceof HubConnectorUpdaterNoPermissionException)
					{
						throw new HubConnectorUpdaterDownloadFailedException(sprintf($this->textPhrases->get_text('UPDATE_NOT_AVAILABLE'),
						                                                             $update['version']));
					}
					
					$this->messageStack->add_session($e->getMessage(), 'error');
					throw new HubConnectorUpdaterDownloadFailedException();
				}
			}
			
			$this->checkDownloadedUpdates();
			
			return;
		}
		
		foreach($this->updateStatus['body']['updates'] as $update)
		{
			try
			{
				$this->downloadUpdateZip($update);
			}
			catch(Exception $e)
			{
				if($e instanceof HubConnectorUpdaterUpdateNotAvailableException === false
				   && $e instanceof HubConnectorUpdaterMD5HashException === false
				   && $e instanceof HubConnectorUpdaterNoPermissionException === false
				)
				{
					# catch only own exceptions
					throw $e;
				}
				elseif($e instanceof HubConnectorUpdaterNoPermissionException)
				{
					throw new HubConnectorUpdaterDownloadFailedException(sprintf($this->textPhrases->get_text('UPDATE_NOT_AVAILABLE'),
					                                                             $update['version']));
				}
				
				$this->messageStack->add_session($e->getMessage(), 'error');
				throw new HubConnectorUpdaterDownloadFailedException();
			}
		}
		
		$this->checkDownloadedUpdates();
		
		return;
	}
	
	
	/**
	 * Starts the installation of all available downloads. Updates will be downloaded bevore installation, if the
	 * "download before update" option is active.
	 *
	 * @throws \HubConnectorUpdaterDownloadFailedException
	 * @throws \HubConnectorUpdaterNoInitialPointException
	 * @throws \HubConnectorUpdaterNoUpdateAvailableException
	 * @throws \HubConnectorUpdaterUpdateFailedException
	 * @throws \HubConnectorUpdaterUpdatesNotDownloadedException
	 */
	public function installUpdates()
	{
		if(count($this->updateStatus['body']['updates']) === 0)
		{
			throw new HubConnectorUpdaterNoUpdateAvailableException($this->textPhrases->get_text('NO_UPDATES'));
		}
		
		if($this->downloadBeforeUpdate)
		{
			$this->downloadUpdates();
		}
		
		if(count($this->updateStatus['body']['updates']) !== count($this->downloadedUpdates))
		{
			throw new HubConnectorUpdaterUpdatesNotDownloadedException($this->textPhrases->get_text('UPDATES_NOT_DOWNLOADED'));
		}
		
		$alreadyInstalledUpdates = [];
		$somethingToUpdate       = isset($this->updateStatus['body']['updates'][$this->hubConnectorUpdaterVersion]);
		if($somethingToUpdate)
		{
			$update = $this->updateStatus['body']['updates'][$this->hubConnectorUpdaterVersion];
			do
			{
				try
				{
					$this->checkPermission($update['version']);
					$this->backupFiles($update['version']);
					$this->copyFiles($update['version']);
					
					$this->setHubConnectorUpdaterVersion($update['version']);
					$this->messageStack->add_session(sprintf($this->textPhrases->get_text('UPDATE_SUCCESS'),
					                                         $update['version']), 'success');
					$alreadyInstalledUpdates[] = $update['version'];
					
					$somethingToUpdate = isset($this->updateStatus['body']['updates'][$update['version']])
					                     && !in_array($update['version'], $alreadyInstalledUpdates);
					if($somethingToUpdate)
					{
						$update = $this->updateStatus['body']['updates'][$this->hubConnectorUpdaterVersion];
					}
				}
				catch(Exception $e)
				{
					if($e instanceof HubConnectorUpdaterNoPermissionException === false
					   && $e instanceof HubConnectorUpdaterBackupCreationFailedException === false
					   && $e instanceof HubConnectorUpdaterCopyFailedException === false
					)
					{
						# throw all exceptions that doesn't come from updater
						throw $e;
					}
					
					if($e instanceof HubConnectorUpdaterNoPermissionException === false)
					{
						$this->messageStack->add_session($e->getMessage(), 'error');
					}
					
					$this->messageStack->add_session(sprintf($this->textPhrases->get_text('IMPORT_BACKUP'),
					                                         $update['version']), 'info');
					
					$this->restoreBackup($update['version']);
					$this->deleteBackupsDirectory();
					$this->deleteUpdatesDirectory();
					
					throw new HubConnectorUpdaterUpdateFailedException(sprintf($this->textPhrases->get_text('UPDATE_FAILED'),
					                                                           $update['version']));
				}
			}
			while($somethingToUpdate === true);
			
			$this->deleteBackupsDirectory();
			$this->deleteUpdatesDirectory();
		}
		else
		{
			throw new HubConnectorUpdaterNoInitialPointException(sprintf($this->textPhrases->get_text('UPDATE_NO_START'),
			                                                             $this->hubConnectorUpdaterVersion));
		}
	}
	
	
	/**
	 * Checks the update server for new updates. Saves the result internally.
	 *
	 * @throws \HubConnectorUpdaterServerNotAvailableException
	 */
	protected function checkUpdates()
	{
		$this->updateStatus = [];
		$options            = [
			CURLOPT_CONNECTTIMEOUT => 10,
			CURLOPT_POST           => true,
			CURLOPT_POSTFIELDS     => [
				'shopVersion'      => gm_get_conf('INSTALLED_VERSION'),
				'updaterVersion'   => $this->hubConnectorUpdaterVersion,
				'connectorVersion' => $this->hubConnectorVersion,
			],
			CURLOPT_RETURNTRANSFER => true,
			CURLOPT_URL            => $this->updateServer,
		];
		
		$curlHandle = curl_init();
		curl_setopt_array($curlHandle, $options);
		$this->updateStatus['body'] = curl_exec($curlHandle);
		$this->updateStatus['body'] = json_decode($this->updateStatus['body'], true);
		$this->updateStatus['info'] = curl_getinfo($curlHandle);
		curl_close($curlHandle);
		
		if($this->updateStatus['info']['http_code'] !== 200)
		{
			throw new HubConnectorUpdaterServerNotAvailableException($this->textPhrases->get_text('NO_UPDATES'));
		}
	}
	
	
	/**
	 * Checks the updates directory for downloaded updates. Saves the result internally.
	 */
	protected function checkDownloadedUpdates()
	{
		$this->downloadedUpdates = [];
		
		if(file_exists($this->updatesDirectory) && is_dir($this->updatesDirectory)
		   && count($this->updateStatus['body']['updates']) > 0
		)
		{
			foreach($this->updateStatus['body']['updates'] as $update)
			{
				$updateDir = $this->updatesDirectory . DIRECTORY_SEPARATOR . str_replace(' ', '_', $update['version']);
				if(file_exists($updateDir) && is_dir($updateDir))
				{
					$this->downloadedUpdates[] = $update['version'];
				}
			}
		}
	}
	
	
	/**
	 * Sets the hub connector updater version to the given value.
	 *
	 * @param string $version
	 */
	protected function setHubConnectorUpdaterVersion($version)
	{
		gm_set_conf($this->gmKeyVersion, $version);
		$this->hubConnectorUpdaterVersion = $version;
	}
	
	
	/**
	 * Downloads a single update als a zip archiv and unpacks it into in the updates directory.
	 * If the zip extension is not available on the webserver, the downloadUpdateNoZip method will be used as a
	 * fallback.
	 *
	 * @param array $update
	 *
	 * @throws \HubConnectorUpdaterNoPermissionException
	 * @throws \HubConnectorUpdaterMD5HashException
	 */
	protected function downloadUpdateZip($update)
	{
		if(!file_exists($this->updatesDirectory) || !is_dir($this->updatesDirectory))
		{
			$createDir = @mkdir($this->updatesDirectory, 0777, true);
			if($createDir === false)
			{
				$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_PERMISSON_DIR'),
				                                         $this->updatesDirectory), 'error');
				throw new HubConnectorUpdaterNoPermissionException();
			}
		}
		
		$curl = curl_init($update['zip']);
		curl_setopt($curl, CURLOPT_NOBODY, true);
		curl_exec($curl);
		$fileExists = curl_getinfo($curl, CURLINFO_HTTP_CODE) === 200;
		curl_close($curl);
		
		if(!$fileExists)
		{
			$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_ZIPFILE_MISSING'),
			                                         $update['version']), 'notice');
			$this->downloadUpdateNoZip($update);
		}
		else
		{
			$zipFile    = $this->updatesDirectory . DIRECTORY_SEPARATOR . str_replace(' ', '_', $update['version'])
			              . '.zip';
			$createFile = file_put_contents($zipFile, fopen($update['zip'], 'r'));
			
			if($createFile === false)
			{
				$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_PERMISSON_FILE'),
				                                         $zipFile), 'error');
				throw new HubConnectorUpdaterNoPermissionException();
			}
			
			try
			{
				if(isset($update['zip_hash']) && md5_file($zipFile) !== $update['zip_hash'])
				{
					throw new HubConnectorUpdaterMD5HashException(sprintf($this->textPhrases->get_text('ERROR_MD5_HASH'),
					                                                      $zipFile));
				}
				
				$this->unzipUpdate($update);
				@unlink($zipFile);
				
				if($this->downloadSuccessNotification)
				{
					$this->messageStack->add_session(sprintf($this->textPhrases->get_text('UPDATE_DOWNLOADED'),
					                                         $update['version']), 'success');
				}
			}
			catch(Exception $e)
			{
				if($e instanceof HubConnectorUpdaterZipCorruptedException === false
				   && $e instanceof HubConnectorUpdaterMD5HashException === false
				)
				{
					# throw all exceptions that doesn't come from updater
					throw $e;
				}
				
				$this->messageStack->add_session($e->getMessage(), 'notice');
				@unlink($zipFile);
				$this->downloadUpdateNoZip($update);
			}
		}
	}
	
	
	/**
	 * Downloads all update files from a single update into in the updates directory.
	 * This method is a fallback, if the zip extension is not available on the webserver.
	 *
	 * @param array $update
	 *
	 * @throws \HubConnectorUpdaterNoPermissionException
	 * @throws \HubConnectorUpdaterUpdateNotAvailableException
	 * @throws \HubConnectorUpdaterMD5HashException
	 */
	protected function downloadUpdateNoZip($update)
	{
		$updateDir = $this->updatesDirectory . DIRECTORY_SEPARATOR . str_replace(' ', '_', $update['version']);
		
		if(!file_exists($updateDir) || !is_dir($updateDir))
		{
			$createDir = @mkdir($updateDir, 0777, true);
			if($createDir === false)
			{
				$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_PERMISSON_DIR'),
				                                         $updateDir), 'error');
				throw new HubConnectorUpdaterNoPermissionException();
			}
		}
		
		if(count($update['filelist']) === 0)
		{
			$this->deleteDirectory($updateDir);
			$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_FILELIST_EMPTY'),
			                                         $update['version']), 'notice');
			throw new HubConnectorUpdaterUpdateNotAvailableException(sprintf($this->textPhrases->get_text('UPDATE_NOT_AVAILABLE'),
			                                                                 $update['version']));
		}
		
		foreach($update['filelist'] as $updateFile)
		{
			$curl = curl_init($updateFile['source']);
			curl_setopt($curl, CURLOPT_NOBODY, true);
			curl_exec($curl);
			$fileExists = curl_getinfo($curl, CURLINFO_HTTP_CODE) === 200;
			curl_close($curl);
			
			if(!$fileExists)
			{
				$this->deleteDirectory($updateDir);
				throw new HubConnectorUpdaterUpdateNotAvailableException(sprintf($this->textPhrases->get_text('UPDATE_NOT_AVAILABLE'),
				                                                                 $update['version']));
			}
			else
			{
				$updateFileDestination    = $updateDir . DIRECTORY_SEPARATOR . $updateFile['destination'];
				$updateFileDestinationDir = substr($updateFileDestination, 0,
				                                   strrpos($updateFileDestination, DIRECTORY_SEPARATOR));
				
				if(!file_exists($updateFileDestinationDir) || !is_dir($updateFileDestinationDir))
				{
					$createDir = @mkdir($updateFileDestinationDir, 0777, true);
					if($createDir === false)
					{
						$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_PERMISSON_DIR'),
						                                         $updateFileDestinationDir), 'error');
						throw new HubConnectorUpdaterNoPermissionException();
					}
				}
				
				$createFile = file_put_contents($updateFileDestination, fopen($updateFile['source'], 'r'));
				if($createFile === false)
				{
					$this->deleteUpdatesDirectory();
					$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_PERMISSON_FILE'),
					                                         $updateFileDestination), 'error');
					throw new HubConnectorUpdaterNoPermissionException();
				}
				elseif(isset($updateFile['hash']) && md5_file($updateFileDestination) !== $updateFile['hash'])
				{
					$this->deleteUpdatesDirectory();
					throw new HubConnectorUpdaterMD5HashException(sprintf($this->textPhrases->get_text('ERROR_MD5_HASH'),
					                                                      $updateFileDestination));
				}
			}
		}
		
		if($this->downloadSuccessNotification)
		{
			$this->messageStack->add_session(sprintf($this->textPhrases->get_text('UPDATE_DOWNLOADED'),
			                                         $update['version']), 'success');
		}
	}
	
	
	/**
	 * Unpacks the zip archiv of a downloaded update.
	 *
	 * @param array $update
	 *
	 * @throws \HubConnectorUpdaterNoPermissionException
	 * @throws \HubConnectorUpdaterZipCorruptedException
	 */
	protected function unzipUpdate($update)
	{
		$zipName   = str_replace(' ', '_', $update['version']) . '.zip';
		$updateDir = $this->updatesDirectory . DIRECTORY_SEPARATOR . str_replace(' ', '_', $update['version']);
		
		if(file_exists($updateDir) && is_dir($updateDir))
		{
			$deleteDir = $this->deleteDirectory($updateDir);
			
			if($deleteDir === false)
			{
				$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_PERMISSON_DIR'),
				                                         $updateDir), 'error');
				throw new HubConnectorUpdaterNoPermissionException();
			}
		}
		
		$createDir = @mkdir($updateDir, 0777, true);
		if($createDir === false)
		{
			$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_PERMISSON_DIR'), $updateDir),
			                                 'error');
			throw new HubConnectorUpdaterNoPermissionException();
		}
		
		$zip          = new ZipArchive;
		$zipOpen      = @$zip->open($this->updatesDirectory . DIRECTORY_SEPARATOR . $zipName);
		$zipExtracted = @$zip->extractTo($updateDir);
		$zipClosed    = @$zip->close();
		if($zipOpen !== true || $zipExtracted === false || $zipClosed === false)
		{
			$this->deleteDirectory($updateDir);
			throw new HubConnectorUpdaterZipCorruptedException(sprintf($this->textPhrases->get_text('ERROR_ZIP_CORRUPTED'),
			                                                           $update['version']));
		}
	}
	
	
	/**
	 * Checks the permission (creation and writing of files and directories) for an update.
	 *
	 * @param $updateVersion
	 *
	 * @throws \HubConnectorUpdaterNoPermissionException
	 */
	protected function checkPermission($updateVersion)
	{
		$updateDir       = $this->updatesDirectory . DIRECTORY_SEPARATOR . str_replace(' ', '_', $updateVersion)
		                   . DIRECTORY_SEPARATOR;
		$updateFiles     = $this->getDirectoryFiles($updateDir);
		$permissionError = [];
		
		if(count($updateFiles) > 0)
		{
			foreach($updateFiles as $updateFile)
			{
				$shopTestFile = str_replace($updateDir, DIR_FS_CATALOG, $updateFile . '.permission_check');
				$shopTestDir  = substr($shopTestFile, 0, strrpos($shopTestFile, DIRECTORY_SEPARATOR));
				
				if(!file_exists($shopTestDir) || !is_dir($shopTestDir))
				{
					$dirCreated = @mkdir($shopTestDir, 0777, true);
					
					if($dirCreated === false)
					{
						if(!in_array($shopTestDir, $permissionError))
						{
							$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_PERMISSON_DIR'),
							                                         $shopTestDir), 'error');
						}
						$permissionError[] = $shopTestDir;
						continue;
					}
				}
				elseif(file_exists($shopTestDir) && is_dir($shopTestDir) && !is_writeable($shopTestDir))
				{
					if(!in_array($shopTestDir, $permissionError))
					{
						$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_PERMISSON_DIR'),
						                                         $shopTestDir), 'error');
					}
					$permissionError[] = $shopTestDir;
					continue;
				}
				
				$fileOpen    = @fopen($shopTestFile, 'w');
				$fileWritten = @fwrite($fileOpen, 'permission test');
				$fileClosed  = @fclose($fileOpen);
				if($fileOpen === false || $fileWritten === false || $fileClosed === false)
				{
					if(!in_array($shopTestFile, $permissionError))
					{
						$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_PERMISSON_FILE'),
						                                         $shopTestFile), 'error');
					}
					$permissionError[] = $shopTestFile;
					continue;
				}
				
				@unlink($shopTestFile);
			}
		}
		
		if(count($permissionError) > 0)
		{
			print_r($this->messageStack);
			throw new HubConnectorUpdaterNoPermissionException();
		}
	}
	
	
	/**
	 * Creates a backup.
	 *
	 * @param $updateVersion
	 *
	 * @throws \HubConnectorUpdaterBackupCreationFailedException
	 * @throws \HubConnectorUpdaterNoPermissionException
	 */
	protected function backupFiles($updateVersion)
	{
		$this->deleteBackupsDirectory();
		
		$backupDir   = $this->backupsDirectory . DIRECTORY_SEPARATOR . str_replace(' ', '_', $updateVersion)
		               . DIRECTORY_SEPARATOR;
		$updateDir   = $this->updatesDirectory . DIRECTORY_SEPARATOR . str_replace(' ', '_', $updateVersion)
		               . DIRECTORY_SEPARATOR;
		$updateFiles = $this->getDirectoryFiles($updateDir);
		
		if(!file_exists($backupDir) || !is_dir($backupDir))
		{
			$dirCreated = @mkdir($backupDir, 0777, true);
			if($dirCreated === false)
			{
				$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_PERMISSON_DIR'),
				                                         $backupDir), 'error');
				throw new HubConnectorUpdaterNoPermissionException();
			}
		}
		
		if(count($updateFiles) > 0)
		{
			foreach($updateFiles as $updateFile)
			{
				$shopFile      = str_replace($updateDir, DIR_FS_CATALOG, $updateFile);
				$backupFile    = str_replace($updateDir, $backupDir, $updateFile);
				$backupFileDir = substr($backupFile, 0, strrpos($backupFile, DIRECTORY_SEPARATOR));
				
				if(file_exists($shopFile))
				{
					
					if(!file_exists($backupFileDir) || !is_dir($backupFileDir))
					{
						@mkdir($backupFileDir, 0777, true);
					}
					
					$fileCopied = @copy($shopFile, $backupFile);
					
					if(!$fileCopied)
					{
						$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_PERMISSON_FILE'),
						                                         $backupFile), 'error');
						throw new HubConnectorUpdaterNoPermissionException();
					}
					elseif(filesize($shopFile) !== filesize($backupFile))
					{
						throw new HubConnectorUpdaterBackupCreationFailedException(sprintf($this->textPhrases->get_text('ERROR_BACKUP_CREATION_FAILED'),
						                                                                   $shopFile));
					}
				}
			}
		}
	}
	
	
	/**
	 * Restores a backup.
	 *
	 * @param $updateVersion
	 *
	 * @throws \HubConnectorUpdaterBackupRestoreFailedException
	 */
	protected function restoreBackup($updateVersion)
	{
		$backupDir   = $this->backupsDirectory . DIRECTORY_SEPARATOR . str_replace(' ', '_', $updateVersion)
		               . DIRECTORY_SEPARATOR;
		$updateDir   = $this->updatesDirectory . DIRECTORY_SEPARATOR . str_replace(' ', '_', $updateVersion)
		               . DIRECTORY_SEPARATOR;
		$updateFiles = $this->getDirectoryFiles($updateDir);
		
		if(count($updateFiles) > 0)
		{
			foreach($updateFiles as $updateFile)
			{
				$shopFile   = str_replace($updateDir, DIR_FS_CATALOG, $updateFile);
				$backupFile = str_replace($updateDir, $backupDir, $updateFile);
				
				if(file_exists($backupFile) && file_exists($shopFile))
				{
					$deleted = @unlink($shopFile);
					if($deleted === false)
					{
						throw new HubConnectorUpdaterBackupRestoreFailedException(sprintf($this->textPhrases->get_text('ERROR_BACKUP_RESTORE_FAILED'),
						                                                                  $updateVersion));
					}
					
					$fileCopied = @copy($backupFile, $shopFile);
					if(!$fileCopied || filesize($shopFile) !== filesize($backupFile))
					{
						throw new HubConnectorUpdaterBackupRestoreFailedException(sprintf($this->textPhrases->get_text('ERROR_BACKUP_RESTORE_FAILED'),
						                                                                  $updateVersion));
					}
				}
			}
		}
	}
	
	
	/**
	 * Copies the update files of a single update into the shop.
	 *
	 * @param array $updateVersion
	 *
	 * @throws \HubConnectorUpdaterNoPermissionException
	 * @throws \HubConnectorUpdaterCopyFailedException
	 */
	protected function copyFiles($updateVersion)
	{
		$updateDir   = $this->updatesDirectory . DIRECTORY_SEPARATOR . str_replace(' ', '_', $updateVersion)
		               . DIRECTORY_SEPARATOR;
		$updateFiles = $this->getDirectoryFiles($updateDir);
		
		if(count($updateFiles) > 0)
		{
			foreach($updateFiles as $updateFile)
			{
				$shopFile = str_replace($updateDir, DIR_FS_CATALOG, $updateFile);
				$shopDir  = substr($shopFile, 0, strrpos($shopFile, DIRECTORY_SEPARATOR));
				
				if(!file_exists($shopDir) || !is_dir($shopDir))
				{
					@mkdir($shopDir, 0777, true);
				}
				
				$fileCopied = @copy($updateFile, $shopFile);
				if(!$fileCopied)
				{
					$this->messageStack->add_session(sprintf($this->textPhrases->get_text('ERROR_PERMISSON_FILE'),
					                                         $shopFile), 'error');
					throw new HubConnectorUpdaterNoPermissionException();
				}
				elseif(filesize($updateFile) !== filesize($shopFile))
				{
					throw new HubConnectorUpdaterCopyFailedException(sprintf($this->textPhrases->get_text('ERROR_COPY_FAILED'),
					                                                         $updateFile, $shopFile));
				}
			}
		}
	}
	
	
	/**
	 * Returns an array with all files of a directory and his subdirectories.
	 *
	 * @param string $dir
	 *
	 * @return array
	 */
	protected function getDirectoryFiles($dir)
	{
		$return = [];
		$files  = scandir($dir);
		foreach($files as $file)
		{
			if($file !== '.' && $file !== '..')
			{
				$path = realpath($dir . DIRECTORY_SEPARATOR . $file);
				if(!is_dir($path))
				{
					$return[] = $path;
				}
				else
				{
					$return = array_merge($return, $this->getDirectoryFiles($path));
				}
			}
		}
		
		return $return;
	}
	
	
	/**
	 * Deletes a directory. Return true on success and false otherwise.
	 *
	 * @param $dir
	 *
	 * @return bool
	 */
	protected function deleteDirectory($dir)
	{
		if(!file_exists($dir))
		{
			return true;
		}
		
		$files = array_diff(scandir($dir), ['.', '..']);
		foreach($files as $file)
		{
			if(is_dir($dir . DIRECTORY_SEPARATOR . $file))
			{
				$deleted = $this->deleteDirectory($dir . DIRECTORY_SEPARATOR . $file);
				if($deleted === false)
				{
					return false;
				}
			}
			else
			{
				$deleted = @unlink($dir . DIRECTORY_SEPARATOR . $file);
				if($deleted === false)
				{
					return false;
				}
			}
		}
		
		return rmdir($dir);
	}
	
	
	/**
	 * Deletes the backups directory. Return true on success and false otherwise.
	 *
	 * @return bool
	 */
	protected function deleteBackupsDirectory()
	{
		return $this->deleteDirectory($this->backupsDirectory);
	}
	
	
	/**
	 * Deletes the updates directory. Return true on success and false otherwise.
	 *
	 * @return bool
	 */
	protected function deleteUpdatesDirectory()
	{
		return $this->deleteDirectory($this->updatesDirectory);
	}
}